"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const document_1 = require("./document");
const utils_1 = require("../utils");
const signaling_1 = require("@phosphor/signaling");
/**
 * VirtualEditor extends the CodeMirror.Editor interface; its subclasses may either
 * fast-forward any requests to an existing instance of the CodeMirror.Editor
 * (using ES6 Proxy), or implement custom behaviour, allowing for the use of
 * virtual documents representing code in complex entities such as notebooks.
 */
class VirtualEditor {
    constructor(language, path, overrides_registry, foreign_code_extractors) {
        /**
         * Virtual documents update guard.
         */
        this.is_update_in_progress = false;
        this.update_lock = false;
        this.virtual_document = new document_1.VirtualDocument(language, path, overrides_registry, foreign_code_extractors, false);
        this.documents_updated = new signaling_1.Signal(this);
        this.documents_updated.connect(this.on_updated.bind(this));
    }
    /**
     * Once all the foreign documents were refreshed, the unused documents (and their connections)
     * should be terminated if their lifetime has expired.
     */
    on_updated(editor, root_document) {
        try {
            root_document.close_expired_documents();
        }
        catch (e) {
            console.warn('LSP: Failed to close expired documents');
        }
    }
    transform_virtual_to_editor(position) {
        return this.virtual_document.transform_virtual_to_editor(position);
    }
    can_update() {
        return !this.is_update_in_progress && !this.update_lock;
    }
    /**
     * Execute provided callback within an update-locked context, which guarantees that:
     *  - the previous updates must have finished before the callback call, and
     *  - no update will happen when executing the callback
     * @param fn - the callback to execute in update lock
     */
    with_update_lock(fn) {
        return __awaiter(this, void 0, void 0, function* () {
            yield utils_1.until_ready(() => this.can_update(), 10, 10).then(() => {
                try {
                    this.update_lock = true;
                    fn();
                }
                finally {
                    this.update_lock = false;
                }
            });
        });
    }
    /**
     * Update all the virtual documents, emit documents updated with root document if succeeded,
     * and resolve a void promise. The promise does not contain the text value of the root document,
     * as to avoid an easy trap of ignoring the changes in the virtual documents.
     */
    update_documents() {
        return __awaiter(this, void 0, void 0, function* () {
            return new Promise((resolve, reject) => __awaiter(this, void 0, void 0, function* () {
                // defer the update by up to 50 ms (10 retrials * 5 ms break),
                // awaiting for the previous update to complete.
                yield utils_1.until_ready(() => this.can_update(), 10, 5).then(() => {
                    try {
                        this.is_update_in_progress = true;
                        this.perform_documents_update();
                        this.documents_updated.emit(this.virtual_document);
                        this.virtual_document.maybe_emit_changed();
                        resolve();
                    }
                    catch (e) {
                        console.warn('Documents update failed:', e);
                        reject(e);
                    }
                    finally {
                        this.is_update_in_progress = false;
                    }
                });
            }));
        });
    }
    // TODO .root is not really needed as we are in editor now...
    document_at_root_position(position) {
        let root_as_source = position;
        return this.virtual_document.root.document_at_source_position(root_as_source);
    }
    root_position_to_virtual_position(position) {
        let root_as_source = position;
        return this.virtual_document.root.virtual_position_at_document(root_as_source);
    }
    get_editor_at_root_position(root_position) {
        return this.virtual_document.root.get_editor_at_source_line(root_position);
    }
    root_position_to_editor_position(root_position) {
        return this.virtual_document.root.transform_source_to_editor(root_position);
    }
}
exports.VirtualEditor = VirtualEditor;
